/*
 * Based on Obdev's AVRUSB code and under the same license.
 *
 * TODO: Make a proper file header. :-)
 */

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <string.h>

#include "Print.h"
#include "UsbKeyboard.h"

#include "usbdrv.h"

// TODO: Work around Arduino 12 issues better.
//#include <WConstants.h>
//#undef int()

#define BUFFER_SIZE 4 // Minimum of 2: 1 for modifiers + 1 for keystroke 

//private: TODO: Make friend?
uchar    reportBuffer[4];    // buffer for HID reports [ 1 modifier byte + (len-1) key strokes]

static uchar    idleRate;           // in 4 ms units 


/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and but we do allow
 * simultaneous key presses. 
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */
PROGMEM char usbHidReportDescriptor[35] = { /* USB report descriptor */
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop) 
  0x09, 0x06,                    // USAGE (Keyboard) 
  0xa1, 0x01,                    // COLLECTION (Application) 
  0x05, 0x07,                    //   USAGE_PAGE (Keyboard) 
  0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl) 
  0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI) 
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0) 
  0x25, 0x01,                    //   LOGICAL_MAXIMUM (1) 
  0x75, 0x01,                    //   REPORT_SIZE (1) 
  0x95, 0x08,                    //   REPORT_COUNT (8) 
  0x81, 0x02,                    //   INPUT (Data,Var,Abs) 
  0x95, BUFFER_SIZE-1,           //   REPORT_COUNT (simultaneous keystrokes) 
  0x75, 0x08,                    //   REPORT_SIZE (8) 
  0x25, 0x65,                    //   LOGICAL_MAXIMUM (101) 
  0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated)) 
  0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application) 
  0x81, 0x00,                    //   INPUT (Data,Ary,Abs) 
  0xc0                           // END_COLLECTION 
};

PROGMEM prog_uint16_t asciiMap[] = {
  KEY_SPACE, KEY_1|SHIFT, KEY_QUOTE|SHIFT, KEY_3|SHIFT, KEY_4|SHIFT,
  KEY_5|SHIFT, KEY_7|SHIFT, KEY_QUOTE|SHIFT, KEY_9|SHIFT, KEY_0|SHIFT,
  KEY_8|SHIFT, KEY_EQUALS|SHIFT, KEY_COMMA, KEY_MINUS, KEY_DOT, KEY_SLASH,
  KEY_0, KEY_1, KEY_2, KEY_3, KEY_4, KEY_5, KEY_6,
  KEY_7, KEY_8, KEY_9, KEY_COLON, KEY_COLON|SHIFT,
  KEY_COMMA|SHIFT, KEY_EQUALS, KEY_DOT|SHIFT, KEY_SLASH|SHIFT,
  KEY_2|SHIFT,
  KEY_A, KEY_B, KEY_C, KEY_D, KEY_E, KEY_F,
  KEY_G, KEY_H, KEY_I, KEY_J, KEY_K, KEY_L, KEY_M,
  KEY_N, KEY_O, KEY_P, KEY_Q, KEY_R, KEY_S, KEY_T,
  KEY_U, KEY_V, KEY_W, KEY_X, KEY_Y, KEY_Z

};



UsbKeyboardDevice::UsbKeyboardDevice () {
    PORTD = 0; // TODO: Only for USB pins?
    DDRD |= ~USBMASK;

#if defined(__AVR_ATmega8__)
    TIMSK&=!(1<<TOIE0);
    //TIMSK|=(1<<TOV1);
# else
    TIMSK0&=!(1<<TOIE0);
    //TIMSK1|=(1<<TOIE1);
#endif

    cli();
    usbDeviceDisconnect();
    usbDeviceConnect();

    usbInit();
      
    sei();

    // TODO: Remove the next two lines once we fix
    //       missing first keystroke bug properly.
    memset(reportBuffer, 0, sizeof(reportBuffer));      
    usbSetInterrupt(reportBuffer, sizeof(reportBuffer));

}

  void UsbKeyboardDevice::update() {
    usbPoll();
  }

 size_t UsbKeyboardDevice::write(uint8_t c) {
 int key;
    if (c == 0xA)
      sendKeyStroke(KEY_ENTER);
    else if (c == 0x8)
      sendKeyStroke(KEY_BACKSPACE);
    else if ('A' <= c && 'Z' >= c)
      sendKeyStroke(pgm_read_word_near(asciiMap + c-0x20), MOD_SHIFT_LEFT);
    else if ('a' <= c && 'z' >= c)
      sendKeyStroke(pgm_read_word_near(asciiMap + c-0x40));
    else if (c >= ' ') {
      key = pgm_read_word_near(asciiMap + c-' ');
      if (key > 0xFF)
        sendKeyStroke(key-0x100, MOD_SHIFT_LEFT);
      else
        sendKeyStroke(key);
    }
    update();
    return 1;
  }
    
  void UsbKeyboardDevice::sendKeyStroke(byte keyStroke) {
    sendKeyStroke(keyStroke, 0);
  }

  void UsbKeyboardDevice::sendKeyStroke(byte keyStroke, byte modifiers) {
      
    while (!usbInterruptIsReady()) {
      // Note: We wait until we can send keystroke
      //       so we know the previous keystroke was
      //       sent.
    }
      
    memset(reportBuffer, 0, sizeof(reportBuffer));

    reportBuffer[0] = modifiers;
    reportBuffer[1] = keyStroke;
        
    usbSetInterrupt(reportBuffer, sizeof(reportBuffer));

    while (!usbInterruptIsReady()) {
      // Note: We wait until we can send keystroke
      //       so we know the previous keystroke was
      //       sent.
    }
      
    // This stops endlessly repeating keystrokes:
    memset(reportBuffer, 0, sizeof(reportBuffer));      
    usbSetInterrupt(reportBuffer, sizeof(reportBuffer));

  }
    
UsbKeyboardDevice UsbKeyboard = UsbKeyboardDevice();

#ifdef __cplusplus
extern "C"{
#endif 
  // USB_PUBLIC uchar usbFunctionSetup
uchar usbFunctionSetup(uchar data[8]) 
  {
    usbRequest_t    *rq = (usbRequest_t *)((void *)data);

    usbMsgPtr = reportBuffer; //
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){
      /* class request type */

      if(rq->bRequest == USBRQ_HID_GET_REPORT){
	/* wValue: ReportType (highbyte), ReportID (lowbyte) */

	/* we only have one report type, so don't look at wValue */
        // TODO: Ensure it's okay not to return anything here?    
	return 0;

      }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
	//            usbMsgPtr = &idleRate;
	//            return 1;
	return 0;
      }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
	idleRate = rq->wValue.bytes[1];
      }
    }else{
      /* no vendor specific requests implemented */
    }
    return 0;
  }
#ifdef __cplusplus
} // extern "C"
#endif

